@@ -6,7 +6,7 @@
 ;; Maintainer: Protesilaos Stavrou General Issues <~protesilaos/general-issues@lists.sr.ht>
 ;; URL: https://git.sr.ht/~protesilaos/substitute
 ;; Mailing-List: https://lists.sr.ht/~protesilaos/general-issues
-;; Version: 0.1.2
+;; Version: 0.1.4
 ;; Package-Requires: ((emacs "27.1"))
 
 ;; This file is NOT part of GNU Emacs.
@@ -58,6 +58,14 @@ For a reference function, see `substitute-report-operation'."
   :group 'substitute
   :type 'hook)
 
+(defface substitute-match
+  `((t :inherit ,(if-let* ((face 'lazy-highlight)
+                           (facep face))
+                     face
+                   'secondary-selection)))
+  "Face to highlight matches of the given target."
+  :group 'substitute)
+
 (defvar substitute--history '()
   "Minibuffer history for substitution commands.")
 
@@ -82,16 +90,21 @@ Use this to produce a more readable version of TARGET for prompts
 and related."
   (replace-regexp-in-string "\\\\_<\\(?1:.*?\\)\\\\_>" "\\1" target))
 
-(defun substitute--determine-face ()
-  "Return face to highlight target of substitute."
-  (if-let* ((face 'lazy-highlight)
-            (facep face))
-      face
-    'secondary-selection))
+(defun substitute--remove-highlights ()
+  "Remove `substitute-match' overlays."
+  (remove-overlays nil nil 'face 'substitute-match))
+
+(defun substitute--add-highlight (beg end)
+  "Add overlay of `substitute-match' between BEG and END positions."
+  (goto-char beg)
+  (let ((highlight (make-overlay beg end)))
+    (overlay-put highlight 'priority 100)
+    (overlay-put highlight 'face 'substitute-match)))
 
 (defun substitute--prompt-without-highlight (target scope)
   "Prompt for string while referencing TARGET and SCOPE."
   (let ((pretty-target (substitute--prettify-target-description target)))
+    (substitute--collect-targets target scope)
     (read-string
      (format "Substitute `%s' %s with: "
              (propertize pretty-target 'face 'error)
@@ -107,9 +120,11 @@ Highlight the TARGET's matching occurences per the user option
   (let ((pretty-target (substitute--prettify-target-description target)))
     (unwind-protect
         (progn
-          (highlight-regexp target (substitute--determine-face))
+          (substitute--collect-targets target scope)
+          (substitute--highlight-targets)
           (substitute--prompt-without-highlight pretty-target scope))
-      (unhighlight-regexp target))))
+      (substitute--remove-highlights)
+      (setq-local substitute--last-matches nil))))
 
 (defun substitute--prompt (target scope)
   "Return appropriate prompt based on `substitute-highlight'.
@@ -135,7 +150,7 @@ Pass to it the TARGET and SCOPE arguments."
   (widen)
   (if-let* (((region-active-p))
             (bounds (region-bounds)))
-      (goto-char (caadr bounds))
+      (goto-char (cdar bounds))
     (thing-at-point-looking-at target)
     (goto-char (match-end 0))))
 
@@ -157,19 +172,71 @@ Pass to it the TARGET and SCOPE arguments."
     ('defun (substitute--scope-current-defun))
     (_ (substitute--scope-top-of-buffer))))
 
-(defun substitute--operate (target sub &optional scope)
-  "Substitute TARGET with SUB in SCOPE.
-This is the subroutine of `substitute-target' and related."
-  (let ((search-fn (if (eq scope 'above) 're-search-backward 're-search-forward))
-        count)
+
+(defvar-local substitute--last-matches nil
+  "Alist of the last matching substitution targets.
+Each entry is a list of the symbol and its buffer positions.")
+
+(defun substitute--collect-targets (target scope)
+  "Store occurrences of TARGET in SCOPE in `substitute--last-matches'."
+  (let ((search-fn (if (eq scope 'above) 're-search-backward 're-search-forward)))
+    (setq-local substitute--last-matches nil)
     (save-excursion
       (save-restriction
         (substitute--setup-scope target scope)
         (while (funcall search-fn target nil t)
-          (push (match-string-no-properties 0) count)
-          (replace-match sub nil t))))
+          (push (list (match-string-no-properties 0)
+                      (match-beginning 0)
+                      (match-end 0))
+                substitute--last-matches))))
+    substitute--last-matches))
+
+(defun substitute--beg-end (beg end)
+  "Determine if BEG is smaller than END and return ordered list."
+  (if (< beg end)
+      (list beg end)
+    (list end beg)))
+
+(defun substitute--highlight-targets ()
+  "Highlight `substitute--last-matches'."
+  (when-let ((targets substitute--last-matches))
+    (save-excursion
+      (save-restriction
+        (mapcar (lambda (target)
+                  (substitute--add-highlight (nth 1 target)
+                                             (nth 2 target)))
+                targets)))))
+
+(defun substitute--replace-targets (sub &optional scope)
+  "Replace `substitute--last-matches' target with SUB.
+If optional SCOPE is equal to `above', then adjust for a reverse
+motion."
+  (when-let ((targets substitute--last-matches))
+    (save-excursion
+      (when (listp buffer-undo-list)
+        (push (point) buffer-undo-list))
+      (save-restriction
+        (mapcar (lambda (target)
+                  (let ((ps (substitute--beg-end (nth 1 target) (nth 2 target)))
+                        reverse)
+                    (when (eq scope 'above)
+                      (setq reverse t))
+                    (goto-char (if reverse (cadr ps) (car ps)))
+                    (funcall
+                     (if reverse 're-search-backward 're-search-forward)
+                     (car target))
+                    (replace-match sub)))
+                targets)))))
+
+(defun substitute--operate (target sub &optional scope)
+  "Operate on TARGET with SUB in SCOPE."
+  (let* ((targets (or substitute--last-matches
+                      (substitute--collect-targets target scope)))
+         (count (length targets)))
+    (substitute--replace-targets sub scope)
+    (setq-local substitute--last-matches nil)
     (run-hook-with-args 'substitute-post-replace-hook
-                        target sub (length count)
+                        target sub count
                         (substitute--scope-description scope))))
 
 (defun substitute--determine-target ()
@@ -187,7 +254,7 @@ Report a `user-error' if no target is found."
         (format "\\_<%s\\_>" thing)
       (user-error "No substitution target at point")))))
 
-(defmacro substitute-command (fn doc &optional scope)
+(defmacro substitute-define-substitute-command (fn doc &optional scope)
   "Produce substitute command using FN, DOC, and SCOPE."
   `(defun ,fn (target sub)
      ,(format
@@ -204,19 +271,19 @@ boundaries." doc)
               (substitute--prompt target ,scope))))
      (substitute--operate target sub ,scope)))
 
-;;;###autoload
-(substitute-command
+;;;###autoload (autoload 'substitute-target-in-buffer "substitute")
+(substitute-define-substitute-command
  substitute-target-in-buffer
  "throughout the buffer")
 
-;;;###autoload
-(substitute-command
+;;;###autoload (autoload 'substitute-target-in-defun "substitute")
+(substitute-define-substitute-command
  substitute-target-in-defun
  "in the defun (per `narrow-to-defun')"
  'defun)
 
-;;;###autoload
-(substitute-command
+;;;###autoload (autoload 'substitute-target-below-point "substitute")
+(substitute-define-substitute-command
  substitute-target-below-point
  "to the end of the buffer"
  'below)
@@ -224,8 +291,8 @@ boundaries." doc)
 (defalias 'substitute-target-to-end-of-buffer
   'substitute-target-below-point)
 
-;;;###autoload
-(substitute-command
+;;;###autoload (autoload 'substitute-target-above-point "substitute")
+(substitute-define-substitute-command
  substitute-target-above-point
  "to the beginning of the buffer"
  'above)
